/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Abacus MathML Editor.
 *
 * The Initial Developer of the Original Code is 
 * Alexander J. Vincent <ajvincent@juno.com>.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// Utility function to make sure HTML markup does not mess us up

function convertMathToXML() {
  var aDocument = aTransaction.mathNode.ownerDocument;
  var walker = aDocument.createTreeWalker(aTransaction.mathNode, NodeFilter.SHOW_ELEMENT,
                                          abacusMathMLEditor.prototype.acceptNodeFilter, true);
  do {
    if (walker.currentNode.namespaceURI == null) {
      var aNode = walker.currentNode;
      var bNode = aNode;
      var xmlns = null;
      while (bNode && (!xmlns)) {
        xmlns = bNode.getAttribute("xmlns");
        if (!xmlns) {
          bNode = walker.parentNode();
        }
      }
      if (!xmlns) {
        xmlns = MATH_NS;
      }

      // We're going to replace aNode with an equivalent element possessing a namespace.
      bNode = document.createElementNS(xmlns, aNode.localName.toLowerCase());
      bNode.setAttributeNS(XMLNS_NS, "xmlns", xmlns);
      while (aNode.firstChild) {
        bNode.appendChild(aNode.firstChild);
      }
      while (aNode.hasAttributes()) {
        var aAttr = aNode.attributes.item(0);
        bNode.setAttributeNS(aAttr.namespaceURI, aAttr.nodeName, aAttr.nodeValue);
      }

      if (aNode.parentNode) {
        aNode.parentNode.replaceChild(bNode, aNode);
      }

      if (aTransaction.mathNode == aNode) {
        aTransaction.mathNode = bNode;
      }

      walker.currentNode = bNode;
    }
  } while (walker.nextNode());
}

// special transactions for our abacusMathMLEditor transaction manager

function abacusReplaceNodeTxn(oldNode, newNode) {
  this.oldNode = oldNode;
  this.newNode = newNode;
  this.hasExecuted = false;
  this.valuesArray = null;
  this.oldValue = oldNode.value;
}
abacusReplaceNodeTxn.prototype = new nsITransaction();
abacusReplaceNodeTxn.prototype.doTransaction = function() {
  this.oldNode.parentNode.replaceChild(this.newNode, this.oldNode);
  this.hasExecuted = true;
}
abacusReplaceNodeTxn.prototype.undoTransaction = function() {
  if (this.hasExecuted) {
    this.newNode.parentNode.replaceChild(this.oldNode, this.newNode);
    this.oldNode.value = this.oldValue;
    this.hasExecuted = false;
  }
}

function abacusOverlayTxn(aReplaceNodeArray, aValuesArray) {
  for (var i = 0; i < aReplaceNodeArray.length; i++) {
    var errMsg = "abacusOverlayTxn must receive an array of only abacusReplaceNodeTxn objects!";
    assert(aReplaceNodeArray[i] instanceof abacusReplaceNodeTxn, errMsg, true);
  }
  this.replaceNodeArray = aReplaceNodeArray;
}
abacusOverlayTxn.prototype = new nsITransaction();
abacusOverlayTxn.prototype.toString = function() {
  return "[object abacusOverlayTransaction]";
}
abacusOverlayTxn.prototype.doTransaction = function() {
  try {
    for (var i = 0; i < this.replaceNodeArray.length; i++) {
      this.replaceNodeArray[i].doTransaction();
    }
  }
  catch(e) {
    // Something went very wrong!  Back out the entire set of changes.
    this.undoTransaction();
    throw e;
  }
}
abacusOverlayTxn.prototype.undoTransaction = function() {
  for (var i = 0; i < this.replaceNodeArray.length; i++) {
    this.replaceNodeArray[i].undoTransaction();
  }
}

/* command objects and functions */
const abacusUndo = {
  isCommandDisabled: function() {
    var rv = (editor.txmgr.numberOfUndoItems == 0);
    return rv;
  }
};

const abacusRedo = {
  isCommandDisabled: function() {
    var rv = (editor.txmgr.numberOfRedoItems == 0);
    return rv;
  }
};

const abacusMoveSelectionUp = {
  doCommand: function() {
    editor.masterSelections.moveSelectionUp();
    editor.updateCommandsDisabled();
  },

  isCommandDisabled: function() {
    return !editor.masterSelections.selectedParentIndexNode;
  }
};

const abacusMoveSelectionBack = {
  doCommand: function() {
    editor.masterSelections.moveSelectionBack();
    editor.updateCommandsDisabled();
  },

  isCommandDisabled: function() {
    return (editor.masterSelections.selectedNodesArray.length < 1);
  }
};

const abacusUpdateTemplate = {
  isCommandDisabled: function() {
    return !editor.selectedTemplate;
  }
};

const abacusApplyTemplate = {
  isCommandDisabled: function() {
    var tests = [];
    tests.push(!editor.selectedTemplate);
    tests.push(!editor.masterSelections.selectedBlank);
    tests.push(!editor.overlaySelections.selectedBlank);
    tests.push(tests[0] || (tests[1] || tests[2]));
    return (tests[3]);
  }
};

const abacusFloatMenu = {
  isCommandDisabled: function() {
    return true;
/* This is disabled due to a bug in Mozilla.
   A modal dialog cannot apparently open a modeless dialog.  The latter will always be modal.
*/
//    return !(editor.lastPopupShown);
  }
}

const abacusOk = {
  disabledAttrFilter: {
    acceptNode: function(aNode) {
      if (aNode.hasAttribute("disabled")) {
        return NodeFilter.FILTER_ACCEPT;
      }
      return NodeFilter.FILTER_SKIP;
    }
  },

  isCommandDisabled: function() {
    var fitb = editor.masterSelections.selectedBlank;
    if (!fitb) {
      return true;
    }
    var mathNode = fitb.parentNode;
    if (!mathNode) {
      return true;
    }

/* I've been getting a lot of funny errors in this region.  If the mathNode was null or undefined, 
   it should've aborted by now.  But for some reason, I see "mathNode has no properties" errors.  Is mathNode a DOM node
   any more?

   These errors usually happen after I execute an undo command and try to focus on a textbox.
*/

    try {
      var foo = mathNode.nodeName;
      while ((mathNode.namespaceURI != MATH_NS)||(mathNode.localName != "math")) {
        mathNode = mathNode.parentNode;
      }
    } catch(e) {
      warn(e);
      return true;
    }

    var walker = document.createTreeWalker(mathNode, NodeFilter.SHOW_ELEMENT, this.disabledAttrFilter, true);
    var aNode = walker.nextNode();
    var rv = Boolean(aNode);
    if (rv) { // I don't believe you.  Let's see your cards.
      var bNode = aNode;
      while ((bNode)&&(bNode != mathNode)) {
        bNode = bNode.parentNode;
      }
      rv = assert(Boolean(bNode), "TreeWalker is broken!  aNode was not walker.root nor a descendant of walker.root.");

      /* Here's a case where an assert() call comes in very handy.  There's a bug in my Mozilla build at the time of this
         writing, which this code looks for.  Because I know about the bug, I can write my code to override that bug if 
         necessary.  With ecmaDebug == true, I also throw a JavaScript strict warning to notify the developer of the bug.
         Normally, my assert() calls have a third argument of true, which would force an error. I don't want that here.
         The developer can see by reading the source code and the DOM-2 Traversal W3C Recommendation that the code is 
         testing the TreeWalker's result correctly.

         "The first step to confirming there is a bug in someone else's work is confirming there are no bugs in your own."
         -- Alexander J. Vincent, June 30, 2001

         Of course, when the TreeWalker bug is fixed, developers using Mozilla builds from after the bugfix will never 
         see this strict warning.  So the assert() function's return value gives us a workaround for an otherwise 
         very painful bug.
      */
    }
    return rv;
  }
}

// Timeout functions to force actions to happen when the events are done (useful for repainting, etc.)

function editorInit() {
  watchUserErrors(editor, 'init');
/*
  This is code I was working on for editor.updateRenderedTemplates().

  var mathNode = document.getElementById("urn:abacus:test");
  removeWhitespaceByNamespace(mathNode, MATH_NS);
  mathNodePreparer.prepareMathNode(mathNode);
  var mathApplyNode = document.getElementById("urn:abacus:test:0"); // content branch root node
  var aTemplatesList = abacusTemplateSetGetter.start(mathApplyNode);
  editor.prepareTemplateSetSelectors(mathApplyNode.parentNode, aTemplatesList);
*/
}

function editorUpdateRenderedTemplates() {
  editor.updateRenderedTemplates();
}

function updateEditorTemplate() {
  editor.updateTemplate();
}

function execute_mEditBindings() {
  editor.executeBindings();
}

function abacusKillMarkedNodes() {
  editor.killMarkedNodes();
}

// Start defining our Abacus MathML Editor

function abacusMathMLEditor() {
  this.hasInitialized = false;
  this.templatesTable = new abacusTemplatesTable();
  this.languageSets = {};
  this.nodesMarkedForDeath = [];
  this.replaceNodeArray = [];
  this.masterIndex = -1;
  this.lastVisibleBranchNode = null;
  this.currentTemplate = null;
  this.targetNodes = null;
  this.invalidCount = 0;
  this.selectedTemplate = null;
  this.idPrefix = null;
  this.lastPopupShown = null;
  this.txmgr = Components.classes["@mozilla.org/transactionmanager;1"].createInstance(nsITransactionManager);
}
abacusMathMLEditor.prototype = {
  masterSelections:      null,
  overlaySelections:     null,
  acceptButton:          null,
  controlsBox:           null,
  templatesBox:          null,
  commandsArray:         null,
  languageSetMenu:       null,
  languageSetBar:        null,
  uiGroupMenu:           null,
  uiControlMenuItem:     null,
  templateSetGrid:       null,
  templateSetList:       [],
  templateSetMenuLists:  [],
  templateSetName:       null,
  templateSetIndexTotal: null,
  templateSetIndex:      null,
  dialogboxContainer:    null,

  // NodeFilters

  acceptNodeFilter: {
    acceptNode: function(aNode) {
      return NodeFilter.FILTER_ACCEPT;
    },

    toString: function() {
      return "[object abacusNodeFilter]";
    }
  },

  hasMEditIndexFilter: {
    requiredAncestorNode: null,

    acceptNode: function(aNode) {
      var index = aNode.getAttributeNS(MEDIT_NS, "index");
      if (index == "") {
        return NodeFilter.FILTER_SKIP;
      }
      index *= 1;
      if (this.requiredAncestorNode) {
        var bNode = aNode;
        while ((bNode)&&(bNode != this.requiredAncestorNode)) {
          bNode = bNode.parentNode;
        }
        if (!bNode) {
          return NodeFilter.FILTER_SKIP;     
        }
      }
 
      return NodeFilter.FILTER_ACCEPT;
    },

    toString: function() {
      return "[object abacusNodeFilter]";
    }
  },

  hasMEditIndexButIsNotMEditElementFilter: {
    acceptNode: function(aNode) {
      if (aNode.namespaceURI == MEDIT_NS) {
        return NodeFilter.FILTER_SKIP;
      }
      return editor.hasMEditIndexFilter.acceptNode(aNode);
    }
  },

  mEditIndexFilter: {
    targetIndex: "-1",
    acceptNode: function(aNode) {
      if (aNode.getAttributeNS(MEDIT_NS, "index") == this.targetIndex) {
        return NodeFilter.FILTER_ACCEPT;
      }
      return NodeFilter.FILTER_SKIP;
    }
  },

  // Convenience function to find child elements of an element

  getChildElements: function(rootNode) {
    var walker = rootNode.ownerDocument.createTreeWalker(rootNode, NodeFilter.SHOW_ELEMENT, editor.acceptNodeFilter, true);
    var output = "";
    var rv = [];
    if (walker.firstChild()) {
      rv.push(walker.currentNode)
      while (walker.nextSibling()) {
        rv.push(walker.currentNode);
      }
    }
    return rv;
  },

  // Functions to determine a "template set": the templates and their controls used to generate our current expression
  disableMenuItems: function(evt) {
    var targetMenuList = evt.target.parentNode.parentNode;
    if (targetMenuList.getAttribute("class") != 'abacusTemplateSetSelector') {
      return;
    }
    var index = targetMenuList.selectedIndex;
    if (index == 0) {
      return;
    }
    for (var i = 0; i < editor.templateSetMenuLists.length; i++) {
      var menulist = editor.templateSetMenuLists[i];
      if (menulist == targetMenuList) {
        continue;
      }
      var menuitems = menulist.getElementsByTagNameNS(XUL_NS, "menuitem");
      menuitems[index].setAttribute("disabled", "true");
    }
  },

  enableMenuItems: function(evt) {
    if (evt.target.getAttribute("class") != 'abacusTemplateSetSelector') {
      return;
    }
    var targetMenuList = evt.target.parentNode.parentNode;
    var index = targetMenuList.selectedIndex;
    if (index == 0) {
      return;
    }
    for (var i = 0; i < editor.templateSetMenuLists.length; i++) {
      var menulist = editor.templateSetMenuLists[i];
      if (menulist == targetMenuList) {
        continue;
      }
      var menuitems = menulist.getElementsByTagNameNS(XUL_NS, "menuitem");
      menuitems[index].removeAttribute("disabled");
    }
  },

  prepareTemplateSetSelectors: function(aSemanticsNode, aTemplatesList) {
    // I really don't like the name of this function.
    var errMsg;
    var childElements;
    var branchNode;
    var menulist;
    var i;
    var j;
    var label;
    var aTemplateSetIndex;

    // We only want to call this function ONCE at most.
    errMsg = "editor.templateSetList must be an empty array!";
    assert(this.templateSetList instanceof Array, errMsg, true);
    assert(this.templateSetList.length == 0, errMsg, true);
    this.templateSetList = aTemplatesList;

    i = aSemanticsNode.getAttribute("encoding");
    errMsg = "Expected 'MathML-Content', received '" + i + "'";
    assert(i == "MathML-Content", errMsg, true);

    childElements = this.getChildElements(aSemanticsNode);
    branchNode = childElements[0];
    for (i = 0; i < this.templateSetMenuLists.length; i++) {
      menulist = this.templateSetMenuLists[i];
      menulist.appendItem("(Content)", "MathML-Content");
    }
    for (j = 1; j < childElements.length; j++) {
      branchNode = childElements[j];
      assert(branchNode.namespaceURI == MATH_NS, 
             "Expected MathML namespace URI, received " + branchNode.namespaceURI, true);
      if (branchNode.localName == "annotation") {
        continue;
      }
      assert(branchNode.localName == "annotation-xml",
             "Expected annotation or annotation-xml element, received " + branchNode.localName, true);

      assert(branchNode.hasAttribute("encoding"), "annotation-xml element does not have an encoding attribute!", true);
      label = branchNode.getAttribute("encoding");
      if (label == "MathML-Presentation") {
        label = "MathML";
      }

      assert(branchNode.hasAttributeNS(XML_NS, "lang"), "annotation-xml element does not have a xml:lang attribute!", true);
      label = branchNode.getAttributeNS(XML_NS, "lang") + "/" + label;

      for (i = 0; i < this.templateSetMenuLists.length; i++) {
        menulist = this.templateSetMenuLists[i];
        menulist.appendItem(label, label);
      }      
    }

    for (i = 0; i < aTemplatesList.length; i++) {
      if (typeof aTemplatesList[i] == "undefined") {
        continue;
      }
      this.templateSetIndex.appendItem(i, i);
    }
  },

  updateTemplateSetSelectors: function() {
    var menulist;
    var dialogTitle;
    var aTemplate;
    var controlsBox;
    var aControlList;
    var bControlList;
    var searchCriteria;
    var i;
    var index;

    index = this.templateSetIndex.value;
    if (index) {
      index *= 1;
      this.templateSetName.value       = this.templateSetList[index].template.dialogTitle;
      this.templateSetIndexTotal.value = this.templateSetList[index].count;
    } else {
      this.templateSetName.value       = "";
      this.templateSetIndexTotal.value = "";
    }

    while (this.controlsBox.firstChild) {
      this.controlsBox.removeChild(this.controlsBox.firstChild);
    }

    for (i = 0; i < this.templateSetMenuLists.length; i++) {
      menulist = this.templateSetMenuLists[i];
      if (menulist.value == "") {
        continue;
      }
      dialogTitle = this.templateSetName.value;
      if (menulist.value != "MathML-Content") {
        dialogTitle += "/" + menulist.value + "/0";
      }
      searchCriteria = [];
      function aToString() {
        return this.name + ' = ' + this.value;
      }
      searchCriteria.push({name: 'dialogTitle',  value: dialogTitle,  toString: aToString});
      aTemplate = this.templatesTable.getFirstSearchResult(searchCriteria);
      aControlList = aTemplate.contentNode.nextSibling;
      controlsBox = menulist.parentNode.lastChild;

      while (controlsBox.firstChild != controlsBox.lastChild) {
        controlsBox.removeChild(controlsBox.firstChild);
      }
      if (!aControlList) {
        continue;
      }
      aControlList = aControlList.cloneNode(true);

      if (this.controlsBox.childNodes.length == 0) {
        /* This is necessary because of the abacusBindings.xml#mEdit:fitb-set evalExpr method.
           Because we must keep things synchronized, we rely on this little hack to make the templates actually work
           like they are supposed to.

           Note: Recent changes to abacusBindings.xml#mEdit:fitb-set@evalExpr may make this code unnecessary.
        */
        bControlList = aControlList.cloneNode(true);
        while (bControlList.firstChild) {
          this.controlsBox.appendChild(bControlList.firstChild);
        }
      }

      while (aControlList.firstChild) {
        controlsBox.insertBefore(aControlList.firstChild, controlsBox.lastChild);
      }
    }

    window.setTimeout(editorUpdateRenderedTemplates, 0);
    // This is done to allow the mEdit:control bindings to take effect first.
  },

  updateRenderedTemplates: function() {
    var menulist;
    var dialogTitle;
    var searchCriteria;
    var aTemplate;
    var aMathExpression;
    var aMathBox;
    for (i = 0; i < this.templateSetMenuLists.length; i++) {
      menulist = this.templateSetMenuLists[i];
      if (menulist.value == "") {
        continue;
      }
      dialogTitle = this.templateSetName.value;
      if (menulist.value != "MathML-Content") {
        dialogTitle += "/" + menulist.value + "/0";
      }
      searchCriteria = [];
      function aToString() {
        return this.name + ' = ' + this.value;
      }
      searchCriteria.push({name: 'dialogTitle',  value: dialogTitle,  toString: aToString});
      aTemplate = this.templatesTable.getFirstSearchResult(searchCriteria);
      aMathExpression = aTemplate.contentNode.cloneNode(true);
      aMathBox = menulist.parentNode.lastChild.previousSibling;

      while (aMathBox.firstChild) {
        aMathBox.removeChild(aMathBox.firstChild);
      }

/* CRASH HAPPY!  I don't know why.
      aMathBox.appendChild(aMathExpression);
*/
      throw new UserError("This feature does not work yet!  There is a crash bug here!");
    }
  },

  // Identify the templates table we're supposed to use

  getTemplatesTable: function() {
    if (typeof aRuntimeConfig.templatesTable != "undefined") {
      this.templatesTable = aRuntimeConfig.templatesTable;
    } else if (typeof aRuntimeConfig.filesByName != "undefined") {
      for (i = 0; i < aRuntimeConfig.filesByName.length; i++) {
        var aFileInfo = aRuntimeConfig.filesByName[i];
        this.templatesTable.addFile(aFileInfo.url, aFileInfo.templateFileCode);
      }
    } else {
      this.templatesTable.addFile("chrome://abacus/content/templates/base.xml", abacusTemplateFile.BASE_FILE);
      this.templatesTable.addFile("chrome://abacus/content/templates/math.xml", abacusTemplateFile.MATH_FILE);
    }

    for (var i = 0; i < this.templatesTable.files.length; i++) {
      /* In the Abacus Template Editor, the absence of whitespace nodes in the actual template nodes is harmless.
         They were never removed because I never saw it to be necessary.  Here, removing them is necessary to avoid
         polluting the master document with useless whitespace nodes.
      */
      var aTemplateFile = this.templatesTable.files[i];
      removeWhitespaceByNamespace(aTemplateFile.fileDoc, MATH_NS);
      removeWhitespaceByNamespace(aTemplateFile.fileDoc, XHTML_NS);
      removeWhitespaceByNamespace(aTemplateFile.fileDoc, MEDIT_NS);
    }
  },

  // Preparing menus for the user based on their language choice

  prepareLanguageSetMenu: function() {
    this.uiGroupMenu       = document.getElementById("uiGroupMenu").cloneNode(true);
    this.uiGroupMenu.removeAttribute("id");
    this.uiControlMenuItem = document.getElementById("uiControlMenuItem").cloneNode(true);
    this.uiControlMenuItem.removeAttribute("id");
    var languageSetByFile;
    var languageSetByName;
    var languageSet;
    var menuitem;
    var languageSetPopup;

    for (languageSetByFile in this.templatesTable.userInterfaces) {
      languageSetByName = languageSetByFile.substr(0, languageSetByFile.indexOf(" "));
      if (typeof this.languageSets[languageSetByName] == "undefined") {
        this.languageSets[languageSetByName] = [];
      }
      languageSet = this.languageSets[languageSetByName];
      languageSet = languageSet.concat(this.templatesTable.userInterfaces[languageSetByFile]);
      this.languageSets[languageSetByName] = languageSet;
    }
    languageSetPopup = this.languageSetMenu.firstChild;
    for (languageSetByName in this.languageSets) {
      menuitem = document.createElement("menuitem")
      menuitem.setAttribute("label", languageSetByName);
      menuitem.setAttribute("type", "radio");
      menuitem.setAttribute("name", "languageSetSelected");
      if (languageSetByName == defaultLang + "/MathML") {
        menuitem.setAttribute("checked", "true");
        this.prepareLanguageSetBar(languageSetByName);
      }
      languageSetPopup.appendChild(menuitem);
    }
  },

  prepareLanguageSetBar: function(languageSetByName) {
    // remove all but the first child of languageSetBar
    while (this.languageSetBar.lastChild != this.languageSetBar.firstChild) {
      this.languageSetBar.removeChild(this.languageSetBar.lastChild);
    }

    this.languageSetBar.firstChild.setAttribute("value", languageSetByName);

    var alphaList = [];
    var label = "";

    // create menus
    var languageSet = this.languageSets[languageSetByName];
    for (var i = 0; i < languageSet.length; i++) {
      var uiGroupNode = languageSet[i];
      var menu = this.uiGroupMenu.cloneNode(true);
      var menuseparator = menu.firstChild.firstChild;
      for (var j = 0; j < uiGroupNode.childNodes.length; j++) {
        var uiControlNode = uiGroupNode.childNodes[j];
        var menuitem = this.uiControlMenuItem.cloneNode(true);
        menuitem.uiNode = uiControlNode;
        label = uiControlNode.getAttribute("label");
        menuitem.setAttribute("label", label);
        alphaList.push({item: menuitem.cloneNode(true), name: label});
        menu.firstChild.insertBefore(menuitem, menuseparator);
      }
      menu.setAttribute("label", uiGroupNode.getAttribute("label"));
      this.languageSetBar.appendChild(menu);
    }

    // create alphabetized menu
    menu = this.uiGroupMenu.cloneNode(true);
    menuseparator = menu.firstChild.firstChild;
    alphaList.sort(function(a, b) {
      if (a.name > b.name) {
        return +1;
      }
      if (a.name < b.name) {
        return -1;
      }
      if (a.name == b.name) {
        return 0;
      }
    });
    for (i = 0; i < alphaList.length; i++) {
      menu.firstChild.insertBefore(alphaList[i].item, menuseparator);
    }
    menu.setAttribute("label", "Index");
    this.languageSetBar.appendChild(menu);

    this.languageSetBar.removeAttribute("hidden");

    // make sure we have a semantics branch for this language
    var humanLanguage = languageSetByName.substr(0, languageSetByName.indexOf("/"));
    var markupLanguage = languageSetByName.substr(languageSetByName.indexOf("/") + 1);
    if (markupLanguage == "MathML") {
      markupLanguage = "MathML-Presentation";
    }
    var semanticsNode = aTransaction.mathNode.firstChild;
    var errMsg = "Expected math:semantics, received " + semanticsNode.nodeName;
    assert(semanticsNode.namespaceURI == MATH_NS, errMsg, true);
    assert(semanticsNode.localName == "semantics", errMsg, true);
    var semanticsChildren = this.getChildElements(semanticsNode);
    var semanticsBranchNode = null;
    for (i = 1; i < semanticsChildren.length; i++) {
      semanticsBranchNode = semanticsChildren[i];
      if (semanticsBranchNode.hasAttributeNS(MEDIT_NS, "visibleBranch")) {
        semanticsBranchNode.removeAttributeNS(MEDIT_NS, "visibleBranch");
        break;
      }
    }

    var foundBranch = false;
    for (i = 1; i < semanticsChildren.length; i++) {
      semanticsBranchNode = semanticsChildren[i];
      if ((semanticsBranchNode.namespaceURI != MATH_NS)||(semanticsBranchNode.localName != "annotation-xml")) {
        continue;
      }
      if (semanticsBranchNode.getAttributeNS(XML_NS, "lang") != humanLanguage) {
        continue;
      }
      if (semanticsBranchNode.getAttribute("encoding") != markupLanguage) {
        continue;
      }
      foundBranch = true;
      break;
    }

    if (!foundBranch) {
      semanticsBranchNode = document.createElementNS(MATH_NS, "annotation-xml");
      semanticsBranchNode.setAttributeNS(XML_NS, "xml:lang", humanLanguage);
      semanticsBranchNode.setAttribute("encoding", markupLanguage);
      semanticsNode.appendChild(semanticsBranchNode);
      i = this.recreateTemplatesForBranch(semanticsBranchNode);
      assert(i, "Failed to recreate a semantics branch for " + humanLanguage + "/" + markupLanguage + "!", true);
    }
    semanticsBranchNode.setAttributeNS(MEDIT_NS, "mEdit:visibleBranch", "true");
  },

  // Float a menu
  floatLastPopup: function() {
    assert(this.lastPopupShown, "editor.lastPopupShown == null!", true);
    if (typeof this.lastPopupShown.dialogbox == "undefined") {
      var dialogbox = document.getElementById("floatMenuDialogBox").cloneNode(true);
      dialogbox.setAttribute("title", this.lastPopupShown.getAttribute("label"));
      var items = this.lastPopupShown.getElementsByTagName("menuitem");
      var rows = dialogbox.getElementsByTagName("rows")[0];
      for (var i = 0; i < items.length - 1; i++) {
        if (i % 4 == 0) {
          var row = document.createElement("row");
          rows.appendChild(row);
        }
        var button = document.createElement("button");
        var item = items[i];
        button.setAttribute("label", item.getAttribute("label"));
        row.appendChild(button);
      }
      this.dialogboxContainer.appendChild(dialogbox);
      this.lastPopupShown.dialogbox = dialogbox;
    }
    this.lastPopupShown.dialogbox.openDialog();
  },

  // Create the appropriate templates, complete with values, for a semantics branch.

  recreateTemplatesForBranch: function(aSemanticsBranchNode) {
    var errMsg;
    var dialogTitle;
    var replaceNodeArray;
    var semanticsNode = aSemanticsBranchNode.parentNode;
    var semanticsChildren = this.getChildElements(semanticsNode);
    var templateBranch = null;
    for (var i = 0; i < semanticsChildren.length; i++) {
      if (semanticsChildren[i].getAttribute("encoding") == "Abacus-Template-Info") {
        templateBranch = semanticsChildren[i];
        break;
      }
    }
    assert(templateBranch, "recreateTemplatesForBranch requires a template info branch!", true);
    assert(templateBranch.childNodes.length == 1, "Template info branch must have exactly one child node!", true);
    var zeroIndexNode = templateBranch.firstChild;
    assert(zeroIndexNode.getAttributeNS(MEDIT_NS, "index") == "0", "Root branch must have mEdit:index == 0!", true);

    aSemanticsBranchNode.appendChild(templateBranch.firstChild.cloneNode(true));
    if (templateBranch.firstChild.localName == "fitb") {
      // obviously, <mEdit:fitb mEdit:index="0"/>
      return true;
    }

    errMsg = "Expected mEdit:templateInfo, received " + templateBranch.firstChild.nodeName;
    assert(templateBranch.firstChild.localName == "templateInfo", errMsg, true);

    // Okay, we have some work to do.
    var humanLanguage = aSemanticsBranchNode.getAttributeNS(XML_NS, "lang");
    var markupLanguage = aSemanticsBranchNode.getAttribute("encoding");
    if (markupLanguage == "MathML-Presentation") {
      markupLanguage = "MathML";
    }

    var templateInfoNodes = templateBranch.getElementsByTagNameNS(MEDIT_NS, "templateInfo");
    replaceNodeArray = [];
    for (i = 0; i < templateInfoNodes.length; i++) {
      var templateNode = templateInfoNodes[i];
      dialogTitle  = templateNode.getAttribute("operationName") + "/";
      dialogTitle += templateNode.getAttribute("indexCode") + "/";
      dialogTitle += humanLanguage + "/" + markupLanguage + "/0";
      var searchCriteria = [];
      function aToString() {
        return this.name + ' = ' + this.value;
      }
      searchCriteria.push({name: 'dialogTitle',  value: dialogTitle,  toString: aToString});
      var aTemplate = this.templatesTable.getFirstSearchResult(searchCriteria);
      assert(aTemplate, "We did not get a template for " + dialogTitle + "!", true);

      // create the overlay node and the sub-transaction
      overlayNode = aTemplate.contentNode.cloneNode(true);
      this.masterSelections.prepareOverlayNode(overlayNode); // for setting mEdit:adjustIndex="true" attributes
      replaceNodeArray.push(overlayNode);
    }
    /*
      I return false here because at this point, I do not know exactly how to code for this.  Ideally, we would apply
      each template in order of creation (popping from the overlayNode), and then compare the output of the
      Abacus-Template-Info algorithm to what the branch currently holds.  The comparison would tell us how to adjust the
      mEdit:index values.

      People wishing to correctly implement this may wish to look at chrome://abacus/content/tools/getTemplateSet.js
      for code similar to what I had in mind (at least, the checkNode portion where canSubstitute is true).

    */

    return false;
  },

  // Force a template to re-render whenever the user decides to change the number of arguments for a template.

  updateTemplate: function() {
    var i;

    this.templatesBox.controlList = [];
    for (i = 0; i < this.controlsBox.childNodes.length; i++) {
      var aControlNode = this.controlsBox.childNodes[i];
      this.templatesBox.controlList.push({name: aControlNode.controlName, value: aControlNode.value});
    }

    while (this.templatesBox.firstChild) {
      this.templatesBox.removeChild(this.templatesBox.firstChild);
    }
    this.selectedTemplate = this.currentTemplate.contentNode.cloneNode(true);
    this.templatesBox.appendChild(this.selectedTemplate);
    this.overlaySelections.selectedBlank = null;

    this.updateCommandsDisabled();
  },

  // Display a template selected by the user

  showTemplate: function(uiControlNode) {
    var operationName  = uiControlNode.getAttribute("operationName");
    var indexCode      = uiControlNode.getAttribute("indexCode");
    var humanLanguage  = uiControlNode.parentNode.parentNode.getAttributeNS(XML_NS, "lang");
    var markupLanguage = uiControlNode.parentNode.parentNode.getAttribute("markupLanguage");
    var alternateIndex = uiControlNode.getAttribute("alternateIndex");

    var dialogTitle = [operationName, indexCode, humanLanguage, markupLanguage, alternateIndex].join("/");
    // conveniently, the dialogTitle is a shortcut for finding a specific instance of abacusTemplate.
    var searchCriteria = [];
    function aToString() {
      return this.name + ' = ' + this.value;
    }
    searchCriteria.push({name: 'dialogTitle',  value: dialogTitle,  toString: aToString});
    var aTemplate = this.templatesTable.getFirstSearchResult(searchCriteria);
    if (typeof aTemplate.baseTemplate == "undefined") {
      this.setBaseTemplate(aTemplate);
    }
    
    // show mEdit:control elements
    while (this.controlsBox.firstChild) {
      this.controlsBox.removeChild(this.controlsBox.firstChild);
    }
    for (var i = 0; i < aTemplate.controls.length; i++) {
      var aControl = aTemplate.controls[i];
      this.controlsBox.insertBefore(aControl.node, this.controlsBox.lastChild);
    }
    this.currentTemplate = aTemplate;
    window.setTimeout(updateEditorTemplate, 0); // This is done to allow the mEdit:control bindings to take effect first.
  },

  showTemplate_precondition: function(uiControlNode) {
    errMsg = "Unexpected uiControlNode namespaceURI: " + uiControlNode.namespaceURI;
    assert(uiControlNode.namespaceURI == MEDIT_NS + "#templates", errMsg, true);
    assert(uiControlNode.localName == "uiControl", "Unexpected uiControlNode localName: " + uiControlNode.localName, true);
  },

  // repaint-related functions

  markNodeForDeath: function(aNode) {
    /* This function sets us up to wipe out a node after all the events have taken place,
     * forcing a repaint of the XUL canvas in that node's location.
     * Sad experience and many hours of late-night coding have shown this 
     * to be a necessity for now.
     */
    if (this.nodesMarkedForDeath.length == 0) {
      setTimeout(abacusKillMarkedNodes, 0);
    }
    this.nodesMarkedForDeath.push(aNode);
    return true;
  },

  killMarkedNodes: function() {
    while (this.nodesMarkedForDeath.length > 0) {
      var aNode = this.nodesMarkedForDeath.shift();
      aNode.parentNode.removeChild(aNode);
    }
    return true;
  },

  // apply the template as defined by the user

  overlayTemplate: function() {
    // This is the first part of a two-part process.  The second part is executeBindings().

    var i;
    var j;
    var aControlNode;
    var walker;
    var templateSet;
    var masterMaxIndex;
    var overlayMaxIndex;
    var searchCriteria;
    var foundSelectedBlank;
    var replaceNodeArray;
    var defaultTemplate;
    var zeroIndexNode;
    var branchNode;
    var aTemplate;
    var targetMarkupInfo;
    var targetLanguageInfo;
    var overlayNode;
    var overlayDescendants;
    var fitbSet;
    var descendant;
    var oldIndex;
    var newIndex;
    var templateIndexNodes;
    var valuesArray;
    var aOverlayTxn;
    var aNode;

    // first, establish just which mEdit:index value we're using
    this.masterIndex = this.masterSelections.selectedIndex;

    // get all the nodes we will be creating overlays for
    walker = document.createTreeWalker(this.masterSelections.firstChild, NodeFilter.SHOW_ELEMENT,
                                       this.mEditIndexFilter, true);
    this.mEditIndexFilter.targetIndex = this.masterIndex.toString();
    this.targetNodes = [];
    while (walker.nextNode()) {
      this.targetNodes.push(walker.currentNode);
    }
    assert(this.targetNodes.length > 1, "Ack!  We didn't get enough target nodes for the overlay!", true);

    // narrow down the template set we'll be searching through
    searchCriteria = [];
    searchCriteria.push({name: "operationName",  value: this.currentTemplate.operationName});
    searchCriteria.push({name: "indexCode",      value: this.currentTemplate.indexCode});
    templateSet = this.templatesTable.getMatchingTemplates(searchCriteria);

    masterMaxIndex = this.masterSelections.maxIndex + 1;
    overlayMaxIndex = this.overlaySelections.maxIndex;

    // start creating overlays
    this.mEditIndexFilter.targetIndex = "0" // Reuse the tree walker:  waste not, want not
    foundSelectedBlank = false; // This is a sanity check used later in an assert() call.
    replaceNodeArray = [];
    for (i = 0; i < this.targetNodes.length; i++) {
      // Our goal here is to find a defaultTemplate for each targetNode.
      defaultTemplate = null;
      walker.currentNode = this.targetNodes[i];
      zeroIndexNode = null;
      if (this.mEditIndexFilter.acceptNode(walker.currentNode) == NodeFilter.FILTER_ACCEPT) {
        zeroIndexNode = walker.currentNode;
      } else {
        zeroIndexNode = walker.parentNode();
      }
      assert(zeroIndexNode, "What happened to the node with mEdit:index='0'?", true);

      branchNode = zeroIndexNode.parentNode;
      // the branch node tells us what humanLanguage and markupLanguage attributes to use.
      assert(branchNode.namespaceURI != MEDIT_NS, "There should be no mEdit nodes above the mEdit:index='0' node!", true);

      targetMarkupInfo = branchNode.getAttribute("encoding");

      if (this.targetNodes[i] == this.masterSelections.selectedBlank) {
        // That was easy.
        defaultTemplate = this.currentTemplate;
        foundSelectedBlank = true;
      } else {
        // We're still looking to find a default template for this node.
        searchCriteria = [];
        aTemplate = this.currentTemplate;
        switch (targetMarkupInfo) {
          case "Abacus-Template-Info":
            // special case!  This tells us how to recreate a template later.
            overlayNode = document.createElementNS(MEDIT_NS, "mEdit:templateInfo");
            overlayNode.setAttribute("operationName", aTemplate.operationName);
            overlayNode.setAttribute("indexCode", aTemplate.indexCode);
            overlayNode.setAttributeNS(MEDIT_NS, "mEdit:index", "0");

            var controlsBox = this.controlsBox.cloneNode(true);
            var controlList = document.createElementNS(MEDIT_NS, "mEdit:controlList");
            while (controlsBox.firstChild) {
              var mEditControl = controlsBox.firstChild.getAttribute("name");
              j = 0;
              while (editor.templatesBox.controlList[j].name != mEditControl) {
                j++;
              }
              controlsBox.firstChild.setAttribute("value", editor.templatesBox.controlList[j].value);
              controlList.appendChild(controlsBox.firstChild);
            }
            overlayNode.appendChild(controlList);
            var blanks = this.overlaySelections.firstChild.getElementsByTagNameNS(MEDIT_NS, "fitb");
            for (j = 0; j < blanks.length; j++) {
              overlayNode.appendChild(blanks[j].cloneNode(true));
            }
            break;

          case "MathML-Content": 
          // content MathML fragment.  Use the content template only!
            searchCriteria.push({name: "dialogTitle", value: aTemplate.operationName + "/" + aTemplate.indexCode});
            break;

          case "MathML-Presentation":
            targetMarkupInfo = "MathML";
            // fall through, we still need to get the xml:lang value

          default:
            targetLanguageInfo = getLanguage(branchNode);
            searchCriteria.push({name: "humanLanguage",  value: targetLanguageInfo});
            searchCriteria.push({name: "markupLanguage", value: targetMarkupInfo});
            searchCriteria.push({name: "alternateIndex", value: 0});
            // we only allow the visible template to have an alternateIndex != 0.
        }
        if (targetMarkupInfo != "Abacus-Template-Info") {
          defaultTemplate = this.templatesTable.getFirstSearchResult(searchCriteria, templateSet);
        }
      }

      if (targetMarkupInfo != "Abacus-Template-Info") {
        assert(defaultTemplate, "We didn't get a default template to overlay into the appropriate branch node!", true);

        // create the overlay node and the sub-transaction
        overlayNode = defaultTemplate.contentNode.cloneNode(true);
      }
      this.masterSelections.prepareOverlayNode(overlayNode); // for setting mEdit:adjustIndex="true" attributes
      replaceNodeArray.push(new abacusReplaceNodeTxn(this.targetNodes[i], overlayNode));
    }
    assert(foundSelectedBlank, "We missed the selected blank!", true);

    // move the values of the template into the master expression's indexSynchronizer
    var masterSynch  = this.masterSelections.indexSynchronizer;
    var overlaySynch = this.overlaySelections.indexSynchronizer;
    overlaySynch.setIndexValue(this.overlaySelections.selectedIndex, this.overlaySelections.selectedBlank.value);
    for (i = 1; i <= overlayMaxIndex; i++) {
      var actualIndexValue = overlaySynch.getIndexValue(i);
      masterSynch.setIndexValue(i + masterMaxIndex - 1, actualIndexValue);
    }

    // create and execute the transaction
    aOverlayTxn = new abacusOverlayTxn(replaceNodeArray, valuesArray);
    this.lastValuesArray = valuesArray;
    this.lastMasterMaxIndex = masterMaxIndex;
    this.lastOverlayIndex = this.overlaySelections.selectedIndex;
    this.txmgr.doTransaction(aOverlayTxn);
    setTimeout(execute_mEditBindings, 0);
  },

  executeBindings: function() {
    // This is the second part of a two-part process.  The first part is overlayTemplate().
    var errMsg = "There are still mEdit:fitb-set elements in the document!";
    assert(this.masterSelections.getElementsByTagNameNS(MEDIT_NS, "fitb-set").length == 0, errMsg, true);
    this.masterSelections.adjustIndexes(this.masterIndex, this.lastOverlayIndex, this.lastMasterMaxIndex);
    this.masterIndex = -1; // now that we're done with it
    this.updateCommandsDisabled(null);
  },

  /* Finalize nodes for aTransaction.doTransaction().
   * The idea here is that if the runtime configuration didn't explicitly tell us to return elements with mEdit:index
   * attributes, we cast each one of them out.  (Remember, the mEdit namespace is not necessarily valid outside an Abacus
   * application.  So we expel them like the demons MathML considers them to be.)
   *
   * If the runtime configuration did explicitly request these elements, then we set the value attribute for each index 
   * node.  Thus, when the math node is applied to the master document, the binding won't wipe them out.
   */
  baptizeMathNode: function() {
    var indexNodes = this.masterSelections.getIndexNodes();
    var i;
    var j;
    var aNode;
    var bNode;
    var index;
    var attrName;
    var attrValue;
    var branchCheck;
    var checkArray;
    var walker;

    // clean up obsoleted mEdit:index values and condense
    checkArray = [];
    walker = document.createTreeWalker(aTransaction.mathNode.firstChild, NodeFilter.SHOW_ELEMENT, 
                                       this.mEditIndexFilter, true);
    // the root is our MathML content branch node, <math:semantics/>
    var indexArray = this.masterSelections.indexSynchronizer.indexArray;
    for (i = 0; i < indexArray.length; i++) {
      this.mEditIndexFilter.targetIndex = i.toString();
      walker.currentNode = walker.root;
      if (walker.nextNode()) {
        checkArray.push(i);
      }
    }
    walker = document.createTreeWalker(aTransaction.mathNode, NodeFilter.SHOW_ELEMENT, this.mEditIndexFilter, true);
    for (i = 0; i < checkArray.length; i++) {
      if (checkArray[i] == i) {
        // no sense resetting attributes that haven't changed
        continue;
      }
      walker.currentNode = walker.root;
      this.mEditIndexFilter.targetIndex = checkArray[i].toString();
      while (walker.nextNode()) {
        walker.currentNode.setAttributeNS(MEDIT_NS, "mEdit:index", i);
      }
      j = this.masterSelections.indexSynchronizer.getIndexValue(checkArray[i]);
      this.masterSelections.indexSynchronizer.setIndexValue(i, j);
    }

    if ((typeof aRuntimeConfig.exportIndexNodes != "boolean")||(!aRuntimeConfig.exportIndexNodes)) {
      // export nodes as needed
      for (i = 0; i < indexNodes.length; i++) {
        aNode = indexNodes[i];
        if (aNode.namespaceURI != MEDIT_NS) {
          continue;
        }

        // Are we under <annotation-xml encoding="Abacus-Template-Info"/> ?
        // We have to walk up the tree the hard way, because not every mEdit:foo element has a zeroIndexNode property.
        bNode = aNode;
        branchCheck = false;
        while (!branchCheck) {
          bNode = bNode.parentNode;
          if (bNode.namespaceURI != MATH_NS) {
            continue;
          }
          branchCheck = ((bNode.localName == "semantics")||(bNode.localName == "annotation-xml"));
        }
        if (bNode.getAttribute("encoding") == "Abacus-Template-Info") {
          continue; // we do not want to baptize this... Abacus may edit this expression at a later date
        }

        assert(aNode.localName == "fitb", "Expected mEdit:fitb element, discovered " + aNode.nodeName, true);
        var aRefNode = aNode.exportNode();
        assert(aRefNode, "Failed to receive a reference node for an element with an mEdit:index attribute!", true);
        assert(aRefNode.hasAttributeNS(MEDIT_NS, "index"), "Reference node is missing a mEdit:index attribute!", true);
        if (aRefNode.parentNode) {
          aRefNode.replaceChild(aRefNode.newTextNode, aNode);
        } else {
          aRefNode.appendChild(aRefNode.newTextNode);
          aNode.parentNode.replaceChild(aRefNode, aNode);
        }
      }

      indexNodes = this.masterSelections.getIndexNodes();
      
      // set id or xref attribute as needed
      var idPrefix = this.idPrefix.value;
      aTransaction.mathNode.setAttribute("id", idPrefix);
      for (i = 0; i < indexNodes.length; i++) {
        aNode = indexNodes[i];
        index = aNode.getAttributeNS(MEDIT_NS, "index");
        attrValue = idPrefix + ":" + index;
        bNode = aNode;
        while ((bNode.namespaceURI != MATH_NS)||((bNode.localName != "semantics")&&(bNode.localName != "annotation-xml"))) {
          bNode = bNode.parentNode;
        }
        attrName = (bNode.localName == "semantics") ? "id" : "xref";
        aNode.setAttribute(attrName, attrValue);
        aNode.removeAttributeNS(MEDIT_NS, "index");
      }

      aTransaction.mathNode.removeAttributeNS(XMLNS_NS, "mEdit");
    } else {
      for (i = 0; i < indexNodes.length; i++) {
        indexNodes[i].setAttribute("value", indexNodes[i].value);
      }
    }
  },

  // transaction management 

  updateCommandsDisabled: function(aNode) {
    if (!this.hasInitialized) {
      return;
    }
    for (var i = 0; i < this.commandsArray.length; i++) {
      var aCommand = this.commandsArray[i];
      var disabled = (eval(aCommand.id)).isCommandDisabled(aNode);
      if (aCommand.id == "abacusOk") {
        aCommand = this.acceptButton;
      }
      if (disabled) {
        aCommand.setAttribute("disabled", true);
      } else {
        aCommand.removeAttribute("disabled");
      }
    }
  },

  undoTransaction: function() {
    this.txmgr.undoTransaction();
    var aMaxIndex = this.masterSelections.maxIndex;
    for (var i = 0; i <= aMaxIndex; i++) {
      this.masterSelections.indexSynchronizer.updateValues(i);
    }
    this.updateCommandsDisabled(null);
  },

  redoTransaction: function() {
    this.txmgr.redoTransaction();
    var aMaxIndex = this.masterSelections.maxIndex;
    for (var i = 0; i <= aMaxIndex; i++) {
      this.masterSelections.indexSynchronizer.updateValues(i);
    }
    this.updateCommandsDisabled(null);
  },

  // initialize the editor

  init: function() {
    // pre-emptively disable the accept button until further notice
    this.acceptButton = document.documentElement.getButton('accept');
    this.acceptButton.setAttribute("disabled", "true");

    // set global variables
    this.controlsBox           = document.getElementById("controlsBox");
    this.templatesBox          = document.getElementById("templatesBox");
    this.languageSetBar        = document.getElementById("languageSetBar");
    this.languageSetMenu       = document.getElementById("languageSetMenu");
    this.masterSelections      = document.getElementById("masterSelections");
    this.overlaySelections     = document.getElementById("overlaySelections");
    this.commandsArray         = document.getElementById("abacusMathMLCommands").childNodes;
    this.idPrefix              = document.getElementById("idPrefix");
    this.idPrefix.value        = aTransaction.idPrefix;
/*
  This is code I was working on for editor.updateRenderedTemplates().

    this.templateSetGrid       = document.getElementById("abacusTemplateSetGrid");
    this.templateSetMenuLists  = this.templateSetGrid.getElementsByAttribute("class", "abacusTemplateSetSelector");
    this.templateSetName       = document.getElementById("abacusTemplateSetName");
    this.templateSetIndexTotal = document.getElementById("abacusTemplateSetIndexTotal");
    this.templateSetIndex      = document.getElementById("abacusTemplateSetIndex");

    this.templateSetGrid.addEventListener("command", this.enableMenuItems, true);
    this.templateSetGrid.addEventListener("command", this.disableMenuItems, false);
*/
    this.dialogboxContainer    = document.getElementById("dialogboxContainer");

    var i;

    this.getTemplatesTable();

    this.masterSelections.firstChild.appendChild(aTransaction.mathNode);
    i = mathNodePreparer.prepareMathNode(aTransaction.mathNode);
    assert(i, "Failed to prepare math node!", true);

    if (aTransaction.originalContents instanceof Range) {
      defaultLang = getLanguage(aTransaction.originalContents.commonAncestorContainer);
    } else if (aTransaction.originalContents.nodeType == 1) {
      defaultLang = getLanguage(aTransaction.originalContents);
    } else {
      assert(false, "originalContents must be a DOM Range or Element!", true);
    }

    if (!defaultLang) {
      /* The document didn't tell us what default xml:lang attribute value to use.
       * This means we have to guess at the default language by borrowing from the preferences.
       */
      errMsg = "Abacus MathML Editor should receive a language to use (xml:lang='en-US', lang='en'";
      errMsg+= " or <meta http-equiv='Content-Language' content='en-US'/> for example)!";
      warn(errMsg);

      warn("We need a way to notify the user that we've set a default language they may not want.");
      defaultLang = document.getElementById("bundle_communicator_region").getString("intl.content.langcode");
      aTransaction.mathNode.setAttributeNS(XML_NS, "xml:lang", defaultLang);
    }

    this.prepareLanguageSetMenu();


    this.hasInitialized = true;
    this.updateCommandsDisabled();
  },

  toString: function() {
    return "[object abacusMathMLEditor]";
  }
}
setContract(abacusMathMLEditor.prototype.showTemplate,
            abacusMathMLEditor.prototype.showTemplate_precondition,
            null);

// basic UI functions

function doAssign() {
  // purge our math node of all "sinful" mEdit:foo elements and attributes if necessary
  var rv = watchUserErrors(editor, 'baptizeMathNode');
  aTransaction.mathNode = aTransaction.mathNode.cloneNode(true);
  if (rv) {
    aTransaction.txmgr.doTransaction(aTransaction);
  }
  return rv;
}

function doCancel() {
  return true;
}

ecmaDebug = true;
JS_LIB_DEBUG = true;
var errMsg;

const aTransaction = arguments[0];
const fUtils = new FileUtils();
const parser = new DOMParser();
parser.checkRootElement = function(aDocument, aRootElementName) {
  return (aDocument.documentElement.nodeName == aRootElementName);
}
errMsg = "We didn't get a transaction manager!";
assert(typeof aTransaction.txmgr != "undefined", errMsg, true);
assert(aTransaction.txmgr instanceof Components.interfaces.nsITransactionManager, errMsg, true);
assert(typeof aTransaction.doTransaction == "function", "aTransaction.doTransaction incorrectly defined!", true);
errMsg = "We didn't get a math:math element!";
assert(typeof aTransaction.mathNode != "undefined", errMsg, true);
if (!isXMLDocument(aTransaction.mathNode.ownerDocument)) {
  // HTML mode
  convertMathToXML();
}
assert(aTransaction.mathNode.localName == "math", errMsg, true);
assert(aTransaction.mathNode.namespaceURI == MATH_NS, errMsg, true);
assert(aTransaction.mathNode.nodeType == 1, errMsg, true);
errMsg = "We didn't get an idPrefix string!";
assert(typeof aTransaction.idPrefix == "string", errMsg, true);
const aRuntimeConfig = arguments[1];
var defaultLang = "";
const userErrorBox = document.getElementById("usererrorbox");

// set up our editor
const editor = new abacusMathMLEditor();
window.addEventListener("load", function(evt) {
  setTimeout(editorInit, 0);
}, true);
